#include "pls/internal/scheduling/scheduler.h"
#include "pls/internal/scheduling/static_scheduler_memory.h"
#include "pls/internal/helpers/profiler.h"

using namespace pls::internal::scheduling;

#include <iostream>
#include <complex>
#include <vector>

#include "benchmark_runner.h"
#include "benchmark_base/fft.h"

using namespace comparison_benchmarks::base;

void conquer(fft::complex_vector::iterator data, int n) {
  if (n < 2) {
    return;
  }

  fft::divide(data, n);
  if (n <= fft::RECURSIVE_CUTOFF) {
    fft::conquer(data, n / 2);
    fft::conquer(data + n / 2, n / 2);
  } else {
    scheduler::spawn([data, n]() {
      conquer(data, n / 2);
    });
    scheduler::spawn([data, n]() {
      conquer(data + n / 2, n / 2);
    });
    scheduler::sync();
  }

  fft::combine(data, n);
}

constexpr int MAX_NUM_THREADS = 8;
constexpr int MAX_NUM_TASKS = 32;
constexpr int MAX_STACK_SIZE = 1024 * 4;

static_scheduler_memory<MAX_NUM_THREADS,
                        MAX_NUM_TASKS,
                        MAX_STACK_SIZE> global_scheduler_memory;

int main(int argc, char **argv) {
  int num_threads;
  string directory;
  benchmark_runner::read_args(argc, argv, num_threads, directory);

  string test_name = to_string(num_threads) + ".csv";
  string full_directory = directory + "/PLS_v2/";
  benchmark_runner runner{full_directory, test_name};

  fft::complex_vector data = fft::generate_input();

  scheduler scheduler{global_scheduler_memory, (unsigned) num_threads};

  scheduler.perform_work([&]() {
    for (int i = 0; i < fft::NUM_WARMUP_ITERATIONS; i++) {
      conquer(data.begin(), fft::SIZE);
    }
  });

  scheduler.perform_work([&]() {
    for (int i = 0; i < fft::NUM_ITERATIONS; i++) {
      runner.start_iteration();
      conquer(data.begin(), fft::SIZE);
      runner.end_iteration();
    }
  });
  runner.commit_results(true);

  return 0;
}
