#include "pls/internal/scheduling/scheduler.h"
#include "pls/internal/scheduling/static_scheduler_memory.h"
#include "pls/algorithms/for_each.h"

using namespace pls::internal::scheduling;

#include "benchmark_runner.h"
#include "benchmark_base/matrix.h"

using namespace comparison_benchmarks::base;

template<typename T, int SIZE>
class pls_matrix : public matrix::matrix<T, SIZE> {
 public:
  pls_matrix() : matrix::matrix<T, SIZE>() {}

  void pls_multiply(const matrix::matrix<T, SIZE> &a, const matrix::matrix<T, SIZE> &b) {
    pls::algorithm::for_each_range(0, SIZE, [this, &a, &b](int i) {
      this->multiply_column(i, a, b);
    });
  }
};

constexpr int MAX_NUM_THREADS = 8;
constexpr int MAX_NUM_TASKS = 32;
constexpr int MAX_STACK_SIZE = 1024 * 1;

static_scheduler_memory<MAX_NUM_THREADS,
                        MAX_NUM_TASKS,
                        MAX_STACK_SIZE> global_scheduler_memory;

int main(int argc, char **argv) {
  int num_threads;
  string directory;
  benchmark_runner::read_args(argc, argv, num_threads, directory);

  string test_name = to_string(num_threads) + ".csv";
  string full_directory = directory + "/PLS_v2/";
  benchmark_runner runner{full_directory, test_name};

  pls_matrix<double, matrix::MATRIX_SIZE> a;
  pls_matrix<double, matrix::MATRIX_SIZE> b;
  pls_matrix<double, matrix::MATRIX_SIZE> result;

  scheduler scheduler{global_scheduler_memory, (unsigned) num_threads};

  scheduler.perform_work([&]() {
    for (int i = 0; i < matrix::WARMUP_ITERATIONS; i++) {
      result.pls_multiply(a, b);
    }
  });

  scheduler.perform_work([&]() {
    for (int i = 0; i < matrix::NUM_ITERATIONS; i++) {
      runner.start_iteration();
      result.pls_multiply(a, b);
      runner.end_iteration();
    }
  });
  runner.commit_results(true);
}
